from ckanext.ids.dataspaceconnector.connector import Connector
from ckanext.ids.metadatabroker.client import graphs_to_ckan_result_format

import ckan.plugins as plugins
import ckan.plugins.toolkit as toolkit

import logging
from typing import Set, List, Dict
from ckan.common import config
import requests

import urllib.parse

#logger
log = logging.getLogger("ckanext")

#recomm service setup
recomm_di_host="http://recomm-di"
recomm_sp_host="http://recomm-sp" #config.get("ckanext.ids.trusts_local_dataspace_connector_url")

ckan_url= config.get("ckan.site_url") #config.get("ckanext.ids.trusts_local_dataspace_connector_port")

data_ingestion_port="9090"
service_provider_port="9092"

service_provider_path=recomm_sp_host + ":" + service_provider_port

process_external_url = ckan_url + "/ids/processExternal?uri="

store_interaction_path="/trusts/interaction/store"

interaction_ingestion_url=recomm_di_host + ":" + data_ingestion_port + store_interaction_path

recomm_application_to_user_path="/trusts/reco/ruc2/application-user"
recomm_dataset_to_user_path="/trusts/reco/ruc1/dataset-user"
recomm_service_to_user_path="/trusts/reco/ruc2/service-user"

recomm_application_to_application_path="/trusts/reco/ruc6/application-application"
recomm_application_to_dataset_path="/trusts/reco/ruc4/application-dataset"
recomm_application_to_service_path="/trusts/reco/ruc6/application-service"

recomm_dataset_to_application_path="/trusts/reco/ruc3/dataset-application"
recomm_dataset_to_dataset_path="/trusts/reco/ruc5/dataset-dataset"
recomm_dataset_to_service_path="/trusts/reco/ruc3/dataset-service"

recomm_service_to_application_path="/trusts/reco/ruc6/service-application"
recomm_service_to_dataset_path="/trusts/reco/ruc4/service-dataset"
recomm_service_to_service_path="/trusts/reco/ruc6/service-service"

recomm_dataset_to_user_url=service_provider_path + recomm_dataset_to_user_path
recomm_service_to_user_url=service_provider_path + recomm_service_to_user_path
recomm_application_to_user_url=service_provider_path + recomm_application_to_user_path

recomm_application_to_application_url=service_provider_path + recomm_application_to_application_path
recomm_application_to_dataset_url=service_provider_path + recomm_application_to_dataset_path
recomm_application_to_service_url=service_provider_path + recomm_application_to_service_path

recomm_dataset_to_application_url=service_provider_path + recomm_dataset_to_application_path
recomm_dataset_to_dataset_url=service_provider_path + recomm_dataset_to_dataset_path
recomm_dataset_to_service_url=service_provider_path + recomm_dataset_to_service_path

recomm_service_to_application_url=service_provider_path + recomm_service_to_application_path
recomm_service_to_dataset_url=service_provider_path + recomm_service_to_dataset_path
recomm_service_to_service_url=service_provider_path + recomm_service_to_service_path

#request headers
headers={"Content-type": "application/json", "accept": "application/json;charset=UTF-8"}

#entity types
type_application="application"
type_dataset="dataset"
type_service="service"

#interaction types
download_interaction_type="download"
view_recomm_interaction_type="view_recomm"
publish_interaction_type="publish"
accept_contract_interaction_type="accept_contract"
view_interaction_type="view"

def recomm_recomm_applications_sidebar(
    entity, 
    count):
    
    try:
        userId = plugins.toolkit.g.userobj.id
        
        params = get_recomm_sidebar_params(
            userId,
            entity, 
            count)
            
        url = get_recomm_applications_sidebar_url(
            entity['type'])
            
        if 'userId' not in params:
            recomm_log("Failed to recommended applications for " + entity['type'] + ": " + entity['id'])
            return []
        
        response = requests.get(
            url=url, 
            params=params,
            headers=headers)

        if response.status_code == 200:
            recomm_log("Sucessfully recommended applications for " + entity['type'] + ": " + entity['id'])

            return format_results(response)

        if response.status_code > 200 or response.text is None:
            recomm_log("Failed to recommended applications for " + entity['type'] + ": " + entity['id'])
            return []
        
    except Exception as error:
        recomm_log("Failed to recommended applications")
        return []    

def recomm_recomm_datasets_sidebar(
    entity, 
    count):
    
    try:
        userId = plugins.toolkit.g.userobj.id
        
        params = get_recomm_sidebar_params(
            userId, 
            entity, 
            count)
            
        url = get_recomm_datasets_sidebar_url(
            entity['type'])
            
        if 'userId' not in params:
            recomm_log("Failed to recommended datasets for " + entity['type'] + ": " + entity['id'])
            return []
        
        response = requests.get(
            url=url, 
            params=params,
            headers=headers)

        if response.status_code == 200:
            recomm_log("Sucessfully recommended datasets for " + entity['type'] + ": " + entity['id'])
            
            return format_results(response)

        if response.status_code > 200 or response.text is None:
            recomm_log("Failed to recommended datasets for " + entity['type'] + ": " + entity['id'])
            return []
            
    except Exception as error:
        recomm_log("Failed to recommended datasets")
        return []        
        
def recomm_recomm_services_sidebar(
    entity, 
    count):
    
    try:
        userId = plugins.toolkit.g.userobj.id
        
        params = get_recomm_sidebar_params(
            userId, 
            entity, 
            count)
            
        url = get_recomm_services_sidebar_url(
            entity['type'])
        
        if 'userId' not in params:
            recomm_log("Failed to recommended services for " + entity['type'] + ": " + entity['id'])
            return []
        
        response = requests.get(
            url=url, 
            params=params,
            headers=headers)

        if response.status_code == 200:
            recomm_log("Sucessfully recommended services for " + entity['type'] + ": " + entity['id'])
            
            return format_results(response)

        if response.status_code > 200 or response.text is None:
            recomm_log("Failed to recommended services for " + entity['type'] + ": " + entity['id'])
            return []
            
    except Exception as error:
        recomm_log("Failed to recommended services")
        return []        

def recomm_recomm_datasets_homepage(
    count):
    
    try:
        userId = plugins.toolkit.g.userobj.id
        
        params = {
            "userId": userId, 
            "count": count
        }
        
        response = requests.get(
            url=recomm_dataset_to_user_url, 
            params=params,
            headers=headers)

        if response.status_code == 200:
            recomm_log("Sucessfully recommended datasets for user: " + userId)
            
            return format_results(response)

        if response.status_code > 200 or response.text is None:
            recomm_log("Failed to recommended datasets for user: " + userId)
            return []
    
    except Exception as error:
        recomm_log("Failed to recommended datasets for user")
        return []           

def recomm_recomm_services_homepage(
    count):

    try:
        userId = plugins.toolkit.g.userobj.id
        
        params = {
            "userId": userId, 
            "count": count
        }
        
        response = requests.get(
            url=recomm_service_to_user_url, 
            params=params,
            headers=headers)

        if response.status_code == 200:
            recomm_log("Sucessfully recommended services for user: " + userId)
            
            return format_results(response)

        if response.status_code > 200 or response.text is None:
            recomm_log("Failed to recommended services for user: " + userId)
            return []
            
    except Exception as error:
        recomm_log("Failed to recommended services for user")
        return []       
    
def recomm_recomm_applications_homepage(
    count):
    
    try:
        userId = plugins.toolkit.g.userobj.id
        
        params = {
            "userId": userId, 
            "count": count
        }
        
        response = requests.get(
            url=recomm_application_to_user_url, 
            params=params,
            headers=headers)

        if response.status_code == 200:
            recomm_log("Sucessfully recommended applications for user: " + userId)
            
            return format_results(response)

        if response.status_code > 200 or response.text is None:
            recomm_log("Failed to recommended applications for user: " + userId)
            return []
    
    except Exception as error:
        recomm_log("Failed to recommended applications for user")
        return []    

def recomm_store_download_interaction(
    entityId: str):
    
    try:
        entity = recomm_retrieve_entity(entityId)
        
        if entity is None:
            return False

        data = {
            "entityId": entity["id"],
            "entityType": entity["type"],
            "type": download_interaction_type, 
            "userId": plugins.toolkit.g.userobj.id
        }
    
        response = requests.post(
            url=interaction_ingestion_url,
            json=data,
            headers=headers)

        if response.status_code == 200:
            recomm_log("Sucessfully stored download interaction for: " + entityId)
            return True

        if response.status_code > 200 or response.text is None:
            recomm_log("Failed to store download interaction for: " + entityId)
            return False

    except Exception as error:
        recomm_log("Failed to store download interaction")
        return False
        
def recomm_store_view_recomm_interaction(
    entityId: str, 
    recoId: str):
    
    try:
        entity = recomm_retrieve_entity(entityId)
        
        if entity is None:
            return False

        data = {
            "entityId": entity["id"],
            "entityType": entity["type"],
            "type": view_recomm_interaction_type, 
            "userId": plugins.toolkit.g.userobj.id, 
            "recommenderId": recoId
        }
   
        response = requests.post(
            url=interaction_ingestion_url,
            json=data,
            headers=headers)

        if response.status_code == 200:
            recomm_log("Sucessfully stored view recomm interaction for: " + entityId)
            return True

        if response.status_code > 200 or response.text is None:
            recomm_log("Failed to store view recomm interaction for: " + entityId)
            return False
            
    except Exception as error:
        recomm_log("Failed to store view recomm interaction")
        return False            

def recomm_store_publish_interaction(
    entityId: str, 
    entityType: str):
    
    try:
        data = {
            "entityId": entityId,
            "entityType": entityType,
            "type": publish_interaction_type, 
            "userId": plugins.toolkit.g.userobj.id
        }
        
        response = requests.post(
            url=interaction_ingestion_url,
            json=data,
            headers=headers)

        if response.status_code == 200:
            recomm_log("Sucessfully stored publish interaction for: " + entityId)
            return True

        if response.status_code > 200 or response.text is None:
            recomm_log("Failed to store publish interaction for: " + entityId)
            return False
            
    except Exception as error:
        recomm_log("Failed to store publish interaction")
        return False    

def recomm_store_accept_contract_interaction(
    entityId: str):
    
    try:
        entity = recomm_retrieve_entity(entityId)
        
        if entity is None:
            return False
            
        data = {
            "entityId": entity["id"],
            "entityType": entity["type"],
            "type": accept_contract_interaction_type, 
            "userId": plugins.toolkit.g.userobj.id
        }
        
        response = requests.post(
            url=interaction_ingestion_url,
            json=data,
            headers=headers)

        if response.status_code == 200:
            recomm_log("Sucessfully stored accept contract interaction for: " + entityId)
            return True

        if response.status_code > 200 or response.text is None:
            recomm_log("Failed to store accept contract interaction for: " + entityId)
            return False
            
    except Exception as error:
        recomm_log("Failed to store accept contract interaction")
        return False    
        
def recomm_store_view_interaction(
    entityId: str, 
    entityType: str):
    
    try:
        data = {
            "entityId": entityId, 
            "entityType": entityType,
            "type": view_interaction_type, 
            "userId": plugins.toolkit.g.userobj.id
        }
    
        response = requests.post(
            url=interaction_ingestion_url,
            json=data,
            headers=headers)

        if response.status_code == 200:
            recomm_log("Sucessfully stored view interaction for: " + entityId)
            return True

        if response.status_code > 200 or response.text is None:
            recomm_log("Failed to store view interaction for: " + entityId)
            return False
            
    except Exception as error:
        recomm_log("Failed to store view interaction")
        return False            
        
def recomm_retrieve_entity(
    entityId: str):
    
    local_connector = Connector()
    
    try:
        entityGraphs = local_connector.ask_broker_for_description(element_uri=entityId)
        entity = graphs_to_ckan_result_format(entityGraphs)
    
        return entity

    except Exception as error:
        recomm_log("Failed to retrieve entity: " + entityId)
        return None    

#def recomm_get_interaction_type(
#    entityType: str,
#    interactionType: str):
#    
#    if entityType == type_dataset:
#        interactionType += "_" + type_dataset
#    if entityType == type_service:
#        interactionType += "_" + type_service
#    if entityType == type_application:
#        interactionType += "_" + type_application
#    
#    return interactionType
    
def recomm_log(
    logMessage: str):
    
    log.info("-------------------------");
    log.info("-------------------------");
    log.info("RECOMM | " + logMessage);
    log.info("-------------------------");
    log.info("-------------------------");
    
    return True

def get_recomm_sidebar_params(
    userId:str,
    entity, 
    count):
    
    if(entity['type'] == type_application):
        return {
            "userId": userId, 
            "applicationId": entity['id'],
            "count": count
        }
        
    if(entity['type'] == type_dataset):
        return {
            "userId": userId, 
            "datasetId": entity['id'],
            "count": count
        }
        
    if(entity['type'] == type_service):
        return {
            "userId": userId, 
            "serviceId": entity['id'],
            "count": count
        }  
            
    return {}
    
def get_recomm_applications_sidebar_url(
    entityType:str):
    
    if(entityType == type_application):
        return recomm_application_to_application_url
        
    if(entityType == type_dataset):
        return recomm_application_to_dataset_url
        
    if(entityType == type_service):
        return recomm_application_to_service_url

    return ""

def get_recomm_datasets_sidebar_url(
    entityType:str):
    
    if(entityType == type_application):
        return recomm_dataset_to_application_url
        
    if(entityType == type_dataset):
        return recomm_dataset_to_dataset_url
        
    if(entityType == type_service):
        return recomm_dataset_to_service_url

    return ""

def get_recomm_services_sidebar_url(
    entityType:str):
    
    if(entityType == type_application):
        return recomm_service_to_application_url
        
    if(entityType == type_dataset):
        return recomm_service_to_dataset_url
        
    if(entityType == type_service):
        return recomm_service_to_service_url

    return ""

def format_results(response):

    jsonResponse = response.json()
    entities = []
        
    for result in reversed(jsonResponse["results"]):
        
        result['pkg_id'] = result['id']
        result['id'] = process_external_url + urllib.parse.quote_plus(result['id'])
        result['recoId'] = jsonResponse["reco_id"]
        
        entities.append(result)

    return entities


import json

from six.moves.urllib.parse import urlencode
from urllib3.contrib import pyopenssl
import requests
from requests.exceptions import HTTPError, RequestException
import urllib.parse as urlparse
from urllib.parse import parse_qs
import ckan.plugins.toolkit as toolkit
from ckanext.harvest.harvesters.ckanharvester import CKANHarvester

import logging
log = logging.getLogger(__name__)


class TrustsHarvester(CKANHarvester):
    '''
    A Harvester for CKAN instances. Extends the default from ckan-harvest extension
    '''

    def flatten_query(self, query_list):
        flatten = dict()
        for key, value in query_list:
            flatten[key] = value

        return flatten

    def _get_content(self, url):

        headers = {}
        api_key = self.config.get('api_key')
        if api_key:
            headers['Authorization'] = api_key

        pyopenssl.inject_into_urllib3()
        parsed = urlparse.urlparse(url)
        data = self.flatten_query(urlparse.parse_qsl(parsed.query))
        url_parsed = parsed.scheme + "://" + parsed.netloc + parsed.path
        try:
            http_request = requests.post(url_parsed, json=data, headers=headers)
        except HTTPError as e:
            raise ContentFetchError('HTTP error: %s %s' % (e.response.status_code, e.request.url))
        except RequestException as e:
            raise ContentFetchError('Request error: %s' % e)
        except Exception as e:
            raise ContentFetchError('HTTP general exception: %s' % e)
        return http_request.text

    def _search_for_datasets(self, remote_ckan_base_url, fq_terms=None):
        '''Does a dataset search on a remote CKAN and returns the results.
        Deals with paging to return all the results, not just the first page.
        '''

        base_search_url = remote_ckan_base_url + self._get_search_api_offset()
        params = {'rows': '100', 'start': '0', 'include_private': self.config.get('include_private', 'true')}
        # There is the worry that datasets will be changed whilst we are paging
        # through them.
        # * In SOLR 4.7 there is a cursor, but not using that yet
        #   because few CKANs are running that version yet.
        # * However we sort, then new names added or removed before the current
        #   page would cause existing names on the next page to be missed or
        #   double counted.
        # * Another approach might be to sort by metadata_modified and always
        #   ask for changes since (and including) the date of the last item of
        #   the day before. However if the entire page is of the exact same
        #   time, then you end up in an infinite loop asking for the same page.
        # * We choose a balanced approach of sorting by ID, which means
        #   datasets are only missed if some are removed, which is far less
        #   likely than any being added. If some are missed then it is assumed
        #   they will harvested the next time anyway. When datasets are added,
        #   we are at risk of seeing datasets twice in the paging, so we detect
        #   and remove any duplicates.
        params['sort'] = 'id asc'
        if fq_terms:
            params['fq'] = ' '.join(fq_terms)

        pkg_dicts = []
        pkg_ids = set()
        previous_content = None
        while True:
            url = base_search_url + '?' + urlencode(params)
            log.debug('Searching for CKAN datasets: %s', url)
            try:
                content = self._get_content(url)
            except ContentFetchError as e:
                raise SearchError(
                    'Error sending request to search remote '
                    'CKAN instance %s using URL %r. Error: %s' %
                    (remote_ckan_base_url, url, e))

            if previous_content and content == previous_content:
                raise SearchError('The paging doesn\'t seem to work. URL: %s' %
                                  url)
            try:
                response_dict = json.loads(content)
            except ValueError:
                raise SearchError('Response from remote CKAN was not JSON: %r'
                                  % content)
            try:
                pkg_dicts_page = response_dict.get('result', {}).get('results',
                                                                     [])
            except ValueError:
                raise SearchError('Response JSON did not contain '
                                  'result/results: %r' % response_dict)

            # Weed out any datasets found on previous pages (should datasets be
            # changing while we page)
            ids_in_page = set(p['id'] for p in pkg_dicts_page)
            duplicate_ids = ids_in_page & pkg_ids
            if duplicate_ids:
                pkg_dicts_page = [p for p in pkg_dicts_page
                                  if p['id'] not in duplicate_ids]
            pkg_ids |= ids_in_page

            pkg_dicts.extend(pkg_dicts_page)

            if len(pkg_dicts_page) == 0:
                break

            params['start'] = str(int(params['start']) + int(params['rows']))

        return pkg_dicts

    def modify_package_dict(self, package_dict, harvest_object):

        for index, resource in enumerate(package_dict['resources']):
            package_dict['resources'][index]['url'] = self.transform_url(resource['url'])

        return package_dict

    def transform_url(self, url):
        trusted_connector_port = "8282"
        site_url = toolkit.config.get('ckan.site_url')
        log.info("Transforming url: %s", url)
        log.debug("ckan.site_url is set to: %s", site_url)
        parsed_site_url = urlparse.urlsplit(site_url)
        localhost = parsed_site_url.scheme + "://" + parsed_site_url.hostname + ":" + trusted_connector_port
        log.debug("Local trusted connector is on : %s", localhost)
        resource_path = urlparse.urlsplit(url)
        log.debug("Resource path: %s", resource_path.path)
        # splitting the url based on the ckan.site_url setting
        tranformed_url = localhost + resource_path.path
        log.info("URL is now: %s", tranformed_url)
        return tranformed_url


class ContentFetchError(Exception):
    pass


class ContentNotFoundError(ContentFetchError):
    pass


class RemoteResourceError(Exception):
    pass


class SearchError(Exception):
    pass



import inspect
import json
import logging
import os

import ckan.plugins as plugins
import ckan.plugins.toolkit as toolkit
from ckan.common import config
import ckan.model as model
from ckan.lib.plugins import DefaultTranslation
import ckanext.ids.blueprints as blueprints
import ckanext.ids.swagger as swagger
import ckanext.ids.validator as validator
from ckanext.ids.metadatabroker.client import broker_package_search, strip_scheme
from collections import OrderedDict, Counter

#dtheiler start
from ckanext.ids.recomm.recomm import recomm_recomm_datasets_homepage
from ckanext.ids.recomm.recomm import recomm_recomm_services_homepage
from ckanext.ids.recomm.recomm import recomm_recomm_applications_homepage

from ckanext.ids.recomm.recomm import recomm_recomm_applications_sidebar
from ckanext.ids.recomm.recomm import recomm_recomm_datasets_sidebar
from ckanext.ids.recomm.recomm import recomm_recomm_services_sidebar
#dtheiler end

from ckanext.ids.helpers import check_if_contract_offer_exists, has_more_facets, get_facet_items_dict, string_to_json
from ckanext.scheming.helpers import scheming_get_schema, scheming_field_by_name
from ckanext.vocabularies.helpers import skos_choices_sparql_helper, skos_choices_get_label_by_value

# ToDo make sure this logger is set higher
log = logging.getLogger("ckanext")

#dtheiler start
def recomm_datasets_homepage(count):
    return recomm_recomm_datasets_homepage(count)

def recomm_services_homepage(count):
    return recomm_recomm_services_homepage(count)

def recomm_applications_homepage(count):
    return recomm_recomm_applications_homepage(count)
    
def recomm_applications_sidebar(entity, count):
    return recomm_recomm_applications_sidebar(entity, count)
    
def recomm_datasets_sidebar(entity, count):
    return recomm_recomm_datasets_sidebar(entity, count)

def recomm_services_sidebar(entity, count):
    return recomm_recomm_services_sidebar(entity, count)    
#dtheiler end

class IdsPlugin(plugins.SingletonPlugin, DefaultTranslation):
    log.debug("\n................ Plugin Init 5................\n+")
    plugins.implements(plugins.IConfigurer)
    plugins.implements(plugins.IOrganizationController, inherit=True)
    plugins.implements(plugins.ITranslation)
    plugins.implements(plugins.IValidators)

    plugins.implements(plugins.ITemplateHelpers)

    def get_helpers(self):
        return {
            'ckanext_ids_check_if_contract_offer_exists': check_if_contract_offer_exists,
            'get_facet_items_dict': get_facet_items_dict,
            'has_more_facets': has_more_facets,
            'string_to_json': string_to_json
            }
    
    def get_validators(self):
        return {
            "trusts_url_validator": validator.trusts_url_validator,
            "object_id_validator": validator.object_id_validator,
            "activity_type_exists": validator.activity_type_exists
        }

    from ckanext.ids.model import setup as db_setup

    db_setup()

    # IConfigurer
    def update_config(self, config_):
        toolkit.add_template_directory(config_, 'templates')
        toolkit.add_public_directory(config_, 'public')
        toolkit.add_resource('assets',
                             'ckanext-ids')

    def update_config_schema(self, schema):
        ignore_missing = toolkit.get_validator('ignore_missing')
        is_boolean = toolkit.get_validator('boolean_validator')

        schema.update({
            # This is an existing CKAN core configuration option, we are just
            # making it available to be editable at runtime
            'ckan.search.show_all_types': [ignore_missing, is_boolean],
            'ckanext.ids.connector_catalog_iri': [ignore_missing],
            'ckanext.ids.usage_control_policies': [ignore_missing]

        })

        return schema
    # before rendering organization view
    def before_view(self, organization):
        log.debug("\n................ Before View ................\n+")
        data_application = {
            'fq': '+type:application +organization:' + organization['name'],
            'include_private': True,
            'ext_include_broker_results': False
        }
        application_search = toolkit.get_action("package_search")(None,
                                                                  data_application)
        data_service = {
            'fq': '+type:service +organization:' + organization['name'],
            'include_private': True,
            'ext_include_broker_results': False
        }
        service_search = toolkit.get_action("package_search")(None,
                                                              data_service)
        data_dataset = {
            'fq': '+type:dataset +organization:' + organization['name'],
            'include_private': True,
            'ext_include_broker_results': False
        }
        dataset_search = toolkit.get_action("package_search")(None,
                                                              data_dataset)
        organization['application_count'] = application_search['count']
        organization['service_count'] = service_search['count']
        organization['dataset_count'] = dataset_search['count']
        return organization

    # IBlueprint
    plugins.implements(plugins.IBlueprint)

    def get_blueprint(self):
        return [blueprints.ids_actions]
        

# The following code is an example of how we can implement a plugin that performs an action on a specific event.
# The event is a package read event, show it will be activated whenever a package is read ie. opening the URL of a
# package on the browser. The plugin implements for this the IPackageController. When this is activated, it enqueues a
# background job that will execute the print_test method asynchronously and also executes the same method synchronously.
# If this plugin is enabled, you will see the message 'This is a synchronous test' on the output of the debugging server
# whenever a package is read. The job queue can be seen if you issue the command
# 'ckan -c /etc/ckan/debugging.ini jobs list'
# You can run a worker that will start picking up jobs from the queue list with the command
# 'ckan -c /etc/ckan/debugging.ini jobs worker'
# Then on your terminal you will see the messages produced by the job.


class IdsDummyJobPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IPackageController, inherit=True)

    def read(self, entity):
        toolkit.enqueue_job(blueprints.print_test, [u'This is an async test'])
        blueprints.print_test('This is a synchronous test')

        return entity


def assert_config():
    configuration_keys = {
        'ckanext.ids.trusts_local_dataspace_connector_url',
        'ckanext.ids.trusts_local_dataspace_connector_username',
        'ckanext.ids.trusts_local_dataspace_connector_password',
        'ckan.site_url'
    }
    for key in configuration_keys:
        try:
            assert toolkit.config.get(key) is not None
        except AssertionError:
            raise EnvironmentError(
                'Configuration property {0} was not set. '
                'Please fix your configuration.'.format(
                    key))


# used to load the policy templates from a local json file. A default is already provided.
# For now the name and position of the file is hardcoded
def load_usage_control_policies():
    url = "ckanext.ids:usage_control.json"
    module, file_name = url.split(':', 1)
    try:
        m = __import__(module, fromlist=[''])
    except ImportError:
        return

    p = os.path.join(os.path.dirname(inspect.getfile(m)), file_name)
    if os.path.exists(p):
        with open(p) as schema_file:
            return json.load(schema_file)


def transform_usage_control_policies(policies):
    for policy in policies["policy_templates"]:
        for field in policy["fields"]:
            field["field_name"] = policy["type"] + "_" + field["field_name"]
            try:
                field["form_attrs"]["disabled"] = ""
            except KeyError:
                field["form_attrs"] = {"disabled": ""}


def get_usage_control_policies():
    usage_control_policies = load_usage_control_policies()
    transform_usage_control_policies(usage_control_policies)
    return usage_control_policies


def dictionize_licenses():
    licenses_dict = {}
    for license in toolkit.get_action("license_list")(None, None):
        licenses_dict[license["url"]] = license
    return licenses_dict


class IdsResourcesPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IBlueprint)
    assert_config()
    blueprints.create_or_get_catalog_id()
    # fixing the names in the policy templates and adding some default options
    config.store.update(
        {"ckanext.ids.usage_control_policies": get_usage_control_policies(),
         "ckanext.ids.licenses": dictionize_licenses()})

    def get_blueprint(self):
        return [blueprints.ids, swagger.swaggerui_blueprint]

    plugins.implements(plugins.IFacets, inherit=True)

    # Here we define the facets fields that will be added when a package search is triggered.
    def dataset_facets(self, facets_dict, package_type):
        new_facets_dict = OrderedDict()
        new_facets_dict["license_id"] = plugins.toolkit._("License")
        new_facets_dict["theme"] = plugins.toolkit._("Theme")
        new_facets_dict["TimeFrame"] = plugins.toolkit._("Time Frame")

        return new_facets_dict

    plugins.implements(plugins.IPackageController, inherit=True)

    def after_delete(self, context, pkg_dict):
        package_meta = toolkit.get_action("package_show")(None, {
            "id": pkg_dict["id"]})
        blueprints.delete_from_dataspace_connector(package_meta)

    def before_search(self, search_params):
        return search_params

    def after_search(self, search_results, search_params):
        context = plugins.toolkit.c

        if context.action == "search":
            results_from_broker = self.retrieve_results_from_broker(search_params)
            if len(results_from_broker) > 0:
               search_results["results"] = results_from_broker["results"]
               search_results["count"] = results_from_broker["count"]
               search_results["facets"] = results_from_broker["facets"]
               search_results["search_facets"] = self.retrieve_facet_labels(results_from_broker["search_facets"])
        if context.action == "action":
            results_from_broker = self.retrieve_results_from_broker(search_params)
            if len(results_from_broker) > 0:
                search_results["results"].extend(results_from_broker["results"])
                search_results["count"] += results_from_broker["count"]
                search_results["facets"] = self.merge_facets(search_results["facets"], results_from_broker["facets"])
                search_results["search_facets"] = self.merge_search_facets(search_results["search_facets"], results_from_broker["search_facets"])
        else:
            pass

        if "ext_include_tracking" in search_params["extras"]:
            if toolkit.asbool(search_params["extras"]["ext_include_tracking"]):
                for result in search_results["results"]:
                    result['tracking_summary'] = (
                    model.TrackingSummary.get_for_package(result['id']))

        return search_results

    def retrieve_results_from_broker(self, search_params):
        log.debug("\n................ After Search ................\n+")
        # log.debug("\n\nParams------------------------------------------>")
        # log.debug(json.dumps(search_params, indent=2))
        # log.debug("\n\nResults----------------------------------------->")
        # log.debug(json.dumps(search_results, indent=2))

        start = search_params.get("start", 0)
        limit = search_params.get("rows", 20)
        search_query = search_params.get("q", None)

        # The parameters include organizations, we remove this
        fqset = search_params.get("fq", None)
        if fqset is not None:
            fq2 = []
            for f in fqset:
                fq2.extend([x for x in f.split()
                                     if "+organization" not in x])

            fqset = fq2
            fqset.sort()

        fq = tuple(fqset)

        results_from_broker = broker_package_search(q=search_query,
                                                    fq=fq,
                                                    start_offset=start,
                                                    limit=limit,
                                                    facet_fields=search_params.get("facet.field", None))

        # log.debug(".\n\n\n---BROKER SEARCH RESULTS ARE   ")
        # log.debug(json.dumps([x["name"] for x in  results_from_broker],
        #                     indent=1))
        # log.debug(".\n\n---------------------------:)\n\n ")
        return results_from_broker

    def merge_facets(self, solr_results: dict, broker_results: dict):
        for facet in broker_results.keys():
            if facet in solr_results.keys():
                solr_results[facet] = dict(Counter(broker_results[facet]) + Counter(solr_results[facet]))
            else:
                solr_results[facet] = broker_results[facet]
        return solr_results

    def merge_search_facets(self, solr_results: dict, broker_results: dict):
        for facet in broker_results.keys():
            if facet in solr_results.keys():
                solr_results_facet_item_names = [item['name'] for item in solr_results[facet]["items"]]
                for broker_item in broker_results[facet]["items"]:
                    if broker_item["name"] in solr_results_facet_item_names:
                        index = solr_results_facet_item_names.index(broker_item["name"])
                        solr_results[facet]["items"][index]["count"] += broker_item["count"]
                    else:
                        solr_results[facet]["items"].append(broker_item)
            else:
                solr_results[facet] = broker_results[facet]
        solr_results = self.retrieve_facet_labels(solr_results)
        return solr_results

    def retrieve_facet_labels(self, results: dict):
        dataset_schema = scheming_get_schema("dataset", "dataset", True)
        for facet_key in results:
            if facet_key == "license_id":
                results["license_id"] = self.retrieve_facet_license_labels(results)["license_id"]
            else:
                schema_field = scheming_field_by_name(dataset_schema["dataset_fields"], facet_key)
                if schema_field["choices_helper"] == "skos_vocabulary_helper":
                    skos_choices = skos_choices_sparql_helper(schema_field)
                    for facet_item_index, facet_item in enumerate(results[facet_key]["items"]):
                        display_name = skos_choices_get_label_by_value(skos_choices, facet_item["name"])
                        results[facet_key]["items"][facet_item_index]["display_name"] = display_name

        return results

    def retrieve_facet_license_labels(self, solr_results: dict):
        licenses = config.get("ckanext.ids.licenses")
        for facet_item_index, facet_item in enumerate(solr_results["license_id"]["items"]):
            display_name = None
            if facet_item["name"].find("http") != -1:
                if facet_item["name"].find("cc-zero") > 0 :
                    display_name = "CC0 1.0"
                else:
                    try:
                        display_name = licenses[facet_item["name"]]["title"]
                    except KeyError:
                        display_name = "Old license url"

                    solr_results["license_id"]["items"][facet_item_index]["display_name"] = display_name
        return solr_results


class TrustsRecommenderPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.ITemplateHelpers)
    plugins.implements(plugins.IConfigurer, inherit=True)

    def update_config(self, config_):
        toolkit.add_template_directory(config_, 'templates_recommender')
    def get_helpers(self):
        return {
            'ckanext_ids_recomm_datasets_homepage': recomm_datasets_homepage,
            'ckanext_ids_recomm_services_homepage': recomm_services_homepage,
            'ckanext_ids_recomm_applications_homepage': recomm_applications_homepage,
            'ckanext_ids_recomm_applications_sidebar': recomm_applications_sidebar,
            'ckanext_ids_recomm_datasets_sidebar': recomm_datasets_sidebar,
            'ckanext_ids_recomm_services_sidebar': recomm_services_sidebar
        }
    # IBlueprint
    plugins.implements(plugins.IBlueprint)

    def get_blueprint(self):
        return [blueprints.trusts_recommender]


class TrustsBlockchainPlugin(plugins.SingletonPlugin):
    pass


import copy
import datetime
import json
import logging
from collections import defaultdict
from urllib.parse import urlsplit

import ckan.lib.base as base
import ckan.lib.helpers as h
import ckan.lib.navl.dictization_functions as dict_fns
import ckan.logic as logic
import ckan.model as model
import ckan.plugins as plugins
import ckan.plugins.toolkit as toolkit
import requests
from ckan.common import _, config
from dateutil import tz
from flask import Blueprint, request
from flask import Response, stream_with_context
from werkzeug.datastructures import ImmutableMultiDict

from ckanext.ids.dataspaceconnector.connector import Connector
from ckanext.ids.dataspaceconnector.contract import Contract
from ckanext.ids.dataspaceconnector.offer import Offer
from ckanext.ids.dataspaceconnector.resource import Resource
from ckanext.ids.dataspaceconnector.subscribe import Subscription
from ckanext.ids.metadatabroker.client import graphs_to_artifacts
from ckanext.ids.metadatabroker.client import graphs_to_ckan_result_format
from ckanext.ids.metadatabroker.client import graphs_to_contracts
from ckanext.ids.model import IdsResource, IdsAgreement, IdsSubscription
from ckanext.ids.activity import create_pushed_to_dataspace_connector_activity, create_created_contract_activity

#dtheiler start
from ckanext.ids.recomm.recomm import recomm_store_view_interaction
from ckanext.ids.recomm.recomm import recomm_store_accept_contract_interaction
from ckanext.ids.recomm.recomm import recomm_store_publish_interaction
from ckanext.ids.recomm.recomm import recomm_store_download_interaction
from ckanext.ids.recomm.recomm import recomm_store_view_recomm_interaction

#dtheiler end

# stefan_gind start
from ckanext.ids.smart_contracts.smart_contract_client import (
    make_request_to_smart_contract_api
)
URL_BLOCKCHAIN=config.get("ckanext.ids.blockchain.api_url", 'http://34.77.109.175:8020/api')
send_request_to_smart_contract_api = make_request_to_smart_contract_api(
    URL_BLOCKCHAIN
)
# stefan_gindl end

tuplize_dict = logic.tuplize_dict
clean_dict = logic.clean_dict
parse_params = logic.parse_params
ValidationError = logic.ValidationError

ids = Blueprint(
    'ids',
    __name__
)

ids_actions = Blueprint(
    'ids_actions',
    __name__
)

trusts_recommender = Blueprint(
    'trusts_recommender',
    __name__
)

trusts_blockchain = Blueprint(
    'trusts_blockchain',
    __name__
)

log = logging.getLogger(__name__)

trusts_recommender_plugin_name = "trusts_recommender"
trusts_blockchain_plugin_name = "trusts_blockchain"

def request_contains_mandatory_files():
    return request.files[
               'Deployment file (docker-compose.xml) - mandatory-upload'].filename != ''


@ids.route('/dataset/<id>/resources/create', methods=['POST'])
def create(id):
    # get clean data from the form, data will hold the common meta for all resources
    data = clean_dict(
        dict_fns.unflatten(tuplize_dict(parse_params(request.form))))
    data['package_id'] = id
    to_delete = ['clear_upload', 'save']


    # add fields ending with url in to_delete dictionary
    for field in data:
        if field.endswith('url') and len(field) > 3:
            to_delete.append(field)

    # these are not needed on the schema so we remove them

    for deletable in to_delete:
        del data[deletable]

    resources = []
    for file in request.files:
        # create a new resource dictionary from a copy of the common
        resource = data.copy()
        files = ImmutableMultiDict([(file, request.files[file])])
        # add the file
        resource.update(clean_dict(
            dict_fns.unflatten(tuplize_dict(parse_params(files)))))
        # file input field from the form will have a different name, we need to change this in upload
        resource['upload'] = resource[file]
        resource['name'] = resource[file].filename
        resource['url'] = resource[file].filename
        # get the name of the file to add it as resource_type
        resource['resource_type'] = resource[file].name.split("-upload", 1)[0]
        del resource[file]
        # add the final resource on the table
        resources.append(resource)

    # iterate through all resources and add them on the package
    for resource in resources:
        toolkit.get_action('resource_create')(None, resource)

    revise_package_dict = {
        "match": {
            "name": id
        },
        "update": {
            "state": "active"
        }
    }
    try:
        toolkit.get_action("package_revise")(None, revise_package_dict)
    # unlikely to happen, but just to demonstrate error handling
    except (ValidationError):
        return base.abort(404, _(u'Dataset not found'))
    # redirect tp dataset read page
    return toolkit.redirect_to('dataset.read',
                               id=id)


@ids.route('/dataset/<id>/resources/delete', methods=['DELETE'])
def delete(id):
    return "deleted"


def push_package_task(dataset_dict):
    return push_to_dataspace_connector(dataset_dict)


def push_organization_task(organization_dict):
    action = 'organization_create'
    push_to_central(data=organization_dict, action=action)


def push_to_central(data, action):
    # We'll use the package_create function to create a new dataset.
    node_url = config.get('ckanext.ids.trusts_central_node_ckan')
    url = node_url + action
    # we need to check if the organization exists
    response = requests.post(url, json=data)
    # handle error
    assert response.status_code == 200


def push_to_dataspace_connector(data):
    """
    If data is already in dataspace connector, nothing will be added
    """
    c = plugins.toolkit.g
    context = {'model': model, 'session': model.Session,
               'user': c.user or c.author, 'auth_user_obj': c.userobj,
               }
    local_connector = Connector()
    local_dsc_api = local_connector.get_resource_api()
    # sync calls to the dataspace connector to create the appropriate objects
    # this will be constant.
    # It might cause an error if the connector does not persist it's data. A restart should fix the problem
    catalog = config.get("ckanext.ids.connector_catalog_iri")
    # try to populate this with fields from the package
    offer = Offer(data)

    # If the offer has, for some reason an IRI, but this does not exist in the
    # local dataspace connector, we just create a new IRI for it
    # FIXME: check to merge with code below
    if offer.offer_iri is not None and not \
            local_dsc_api.resource_exists(
                offer.offer_iri):
        offer.offer_iri = None
        for value in data["resources"]:
            rep_iri = value["representation"]
            if not local_dsc_api.resource_exists(rep_iri):
                value["representation"] = None
            art_iri = value["artifact"]
            if not local_dsc_api.resource_exists(art_iri):
                value["artifact"] = None

    if offer.offer_iri is not None:
        log.info("Checking if offer can be updated:" + offer.offer_iri)
        if local_dsc_api.update_offered_resource(
                offer.offer_iri, offer.to_dictionary()):
            log.info("Success! Offer updated!")
            offers = offer.offer_iri
        else:
            # The offer does not exist on the dataspace connector. This might mean that the offer was manually deleted or
            # lost after some restart. The package dictionary contains the iri of the deleted offer so it fails. For now,
            # manually editing the package and the resources is needed, or even deleting the package and create from scratch.
            # We should implement a method to do this through the admin/manage menu
            message = "Offer not found on the Dataspace Connector."
            result = {"pushed" : False, "message": message}
            log.warn(message)
            return result
    else:
        log.info("Offer was not found, creating a new one on the dataspace connector.")
        offers = local_dsc_api.create_offered_resource(
            offer.to_dictionary())
        log.info("Adding resource to catalog.")
        local_dsc_api.add_resource_to_catalog(catalog,
                                              offers)
    # adding resources

    # If this is a service, we must add a new resource which points to the
    # access URL
    extraresources = []
    if offer.access_url is not None:
        newresource = {"service_accessURL": offer.access_url,
                       "description": "service_base_access_url",
                       "resource_type": "service_base_access_url"}
        extraresources.append(newresource)

    for value in data["resources"] + extraresources:
        log.debug(
            "--- CREATING RESOURCE ------\n" + json.dumps(value, indent=1))
        # this has also to run for every resource
        resource = Resource(value)
        if resource.service_accessURL is None:
            internal_resource_url = transform_url_internal_network(
                value["url"])
        else:
            internal_resource_url = resource.service_accessURL
        representation_metadata = {"title": resource.title,
                                   "mediaType": resource.mediaType}
        artifact_metadata = {"accessUrl": internal_resource_url,
                             "title": resource.title,
                             "description": resource.description}
        if resource.representation_iri is None:
            representation = local_dsc_api.create_representation(
                representation_metadata)
        else:
            local_dsc_api.update_representation(
                representation_iri=resource.representation_iri,
                data=representation_metadata)
            representation = resource.representation_iri

        local_dsc_api.add_representation_to_resource(
                offers, representation)

        # The site_url of CKAN is accessible to the whole world, but not to the
        # DSC. This is specially true if the deployment is local and then
        # CKAN is something like localhost:5000 which won't resolve well in
        # the DSC. Thus we re-write the url to take into account the name by
        # which the CKAN is accessible from the DSC.

        if resource.artifact_iri is None:
            artifact = local_dsc_api.create_artifact(
                data=artifact_metadata)
        else:
            local_dsc_api.update_artifact(
                resource.artifact_iri,
                data=artifact_metadata)
            artifact = resource.artifact_iri
            local_dsc_api.add_artifact_to_representation(representation, artifact)
        if "id" in value:
            # add these on the resource meta
            patch_data = {"id": value["id"], "representation": representation,
                          "artifact": artifact}
            logic.action.patch.resource_patch(context, data_dict=patch_data)


    toolkit.get_action("package_patch")(context,
                                        {"id": data["id"],
                                         "catalog_iri": catalog,
                                         "offer_iri": offers})

    contracts = local_dsc_api.get_contracts(offers)

    if contracts["page"]["totalElements"] > 0:
        # push to the broker if the package has a contract
        local_connector.send_resource_to_broker(resource_uri=offer.offer_iri)

        if trusts_recommender_plugin_name in toolkit.g.plugins:
            #dtheiler start
            recomm_store_publish_interaction(
                offer.offer_iri, #entityId
                data["type"]) #entityType
            #dtheiler end
    else:
        message = "This resource doesn't have any contracts, not pushing to broker"
        result = {"pushed" : False, "message": message}
        log.warn(message)
        return result

    create_pushed_to_dataspace_connector_activity(context, data["id"])
    message = "Asset successfully pushed to the TRUSTS Platform"
    result = {"pushed": True, "message": message}
    log.info(message)
    return result


def transform_url_internal_network(url: str,
                                   container_name: str = "local-ckan",
                                   container_port: str = "5000"):
    site_url = str(toolkit.config.get('ckan.site_url'))
    internal_url = "http://" + container_name + ":" + str(container_port)
    if site_url.endswith("/"):
        internal_url += "/"
    return url.replace(site_url, internal_url)


def delete_from_dataspace_connector(data):
    c = plugins.toolkit.g
    context = {'model': model, 'session': model.Session,
               'user': c.user or c.author, 'auth_user_obj': c.userobj,
               }
    local_resource_dataspace_connector = Connector().get_resource_api()
    catalog = config.get("ckanext.ids.connector_catalog_iri")
    offer = Offer(data)
    for value in data["resources"]:
        # this has also to run for every resource
        resource = Resource(value)
        if resource.representation_iri is not None:
            local_resource_dataspace_connector.delete_representation(resource)
        if resource.artifact_iri is not None:
            local_resource_dataspace_connector.delete_artifact(
                resource.artifact_iri, data={})
    if offer.offer_iri is not None:
        local_resource_dataspace_connector.delete_offered_resource(offer)
    return


def transform_url(url):
    site_url = toolkit.config.get('ckan.site_url')
    log.info("Transforming url: %s", url)
    log.debug("ckan.site_url is set to: %s", site_url)
    log.debug(url)
    # splitting the url based on the ckan.site_url setting
    resource_url_part = url.split(site_url, 1)[1]
    transformed_url = toolkit.config.get(
        'ckanext.ids.central_node_connector_url') + toolkit.config.get(
        'ckanext.ids.local_node_name') + "/ckan/5000" + resource_url_part
    log.info("URL is now: %s", transformed_url)
    return transformed_url


@ids_actions.route('/ids/actions/push_package/<id>', methods=['GET'])
def push_package(id):
    package_meta = toolkit.get_action("package_show")(None, {"id": id})
    if config.get("ckanext.ids.blockchain.enabled", "false") == "true":
        push_to_smart_contract_component(package_meta, id)

    # for index, resource in enumerate(package_meta['resources']):
    #    package_meta['resources'][index]['url_type'] = ''
    #    package_meta['resources'][index]['url'] = transform_url(resource['url'])
    # this is the asynchronous task
    # response = toolkit.enqueue_job(push_package_task, [package_meta])
    # this is the synchronous task
    return push_package_task(package_meta)
    # return json.dumps(response.id)


def push_to_smart_contract_component(package_meta, id):
    dict_smart_contracts = {
        "channel": "mychannel",
        "msp": "Org1MSP",
        "orguid": "Org1_appuser",
        # "orguid": package_meta['owner_org'],
        "assetid": id,
        "title": package_meta['title'],
        "size": package_meta['num_resources'],
        "owner": package_meta['owner_org'],
        # "owner": package_meta['owner_org'],
        "value": package_meta['num_tags'],
        "publisher": package_meta['organization'],
        "creator": package_meta['author'],
        "contactPoint": package_meta['author_email'],
        "keyword": "no_equivalent",
        "authorisation": "no_equivalent",
        "dataAccess": package_meta['url'],
        "creationDate": package_meta['metadata_created'],
        "license": package_meta['license_title'],
        "format": "no_equivalent",
        "accessInterface": "no_equivalent",
        "description": package_meta['notes'],
    }
    r = send_request_to_smart_contract_api(
        'secure_create_asset',
        dict_smart_contracts,
    )
    msg_smart_contracts = r.json()['message']
    print(f"Smart contract output: {msg_smart_contracts}")

    # Code for getting all assets, just uncomment. Debugging required,
    # dict_smart_contracts = {
    #     "channel": "mychannel",
    #     "msp": "Org1MSP",
    #     "orguid": "Org1_appuser"
    # }

    # r = send_request_to_smart_contract_api(
    #     'basic_get_all_assets',
    #     dict_smart_contracts,
    # )
    # print(r.json())


# TODO: Remove when AJAX script is in place
@ids_actions.route('/ids/view/push_package/<id>', methods=['GET'])
def push_package_view(id):
    response = push_package(id)
    if response["pushed"]:
        h.flash_success(response["message"])
    else:
        h.flash_error(response["message"])
    return toolkit.redirect_to('dataset.read', id=id)


@ids_actions.route('/ids/actions/push_organization/<id>', methods=['GET'])
def push_organization(id):
    organization_meta = toolkit.get_action("organization_show")(None,
                                                                {"id": id})
    response = toolkit.enqueue_job(push_organization_task, [organization_meta])
    push_organization_task(organization_meta)
    return json.dumps(response.id)


# TODO: Remove when AJAX script is in place
@ids_actions.route('/ids/view/push_organization/<id>', methods=['GET'])
def push_organization_view(id):
    response = push_organization(id)
    h.flash_success(
        _('Object pushed successfully to Central node, jobId: ') + response)
    return toolkit.redirect_to('organization.read', id=id)


@ids_actions.route('/ids/actions/publish/<id>', methods=['POST'])
def publish_action(id):
    local_connector = Connector()
    local_connector_resource_api = local_connector.get_resource_api()
    c = plugins.toolkit.g
    context = {'model': model, 'session': model.Session,
               'user': c.user or c.author, 'auth_user_obj': c.userobj,
               }
    dataset = toolkit.get_action('package_show')(context, {'id': id})

    # If they are trying to create a contract for a package not yet in the DSC
    # this action will push the package. But if the package already exists
    # nothing new will be pushed
    #    push_to_dataspace_connector(dataset)
    #    dataset = toolkit.get_action('package_show')(context, {'id': id})

    c.pkg_dict = dataset
    if isinstance(request.data, Contract):
        contract_meta = request.data
    else:
        contract_meta = Contract(request.get_json())
    # create the contract
    contract_info = {
        "start": contract_meta.contract_start.isoformat(),
        "end": contract_meta.contract_end.isoformat(),
        "title": contract_meta.title,
        "policies": json.dumps(contract_meta.policies)
    }
    contract_id = local_connector_resource_api.create_contract(contract_info)
    # create the rules
    rules = []
    for policy in contract_meta.policies:
        rule = local_connector_resource_api.get_new_policy(policy)
        rule_id = local_connector_resource_api.create_rule({"value": rule})
        rules.append(rule_id)
    # add rules to contract
    local_connector_resource_api.add_rule_to_contract(contract=contract_id,
                                                      rule=rules)
    log.debug("Rules added on contract.")

    try:
        resource_id = dataset["offer_iri"]
        assert resource_id != ""
    except (KeyError, AssertionError):
        log.info("Asset not yet pushed to the local DSC. I will push it now...")
        push_package(id)
        dataset = toolkit.get_action('package_show')(context, {'id': id})
        log.info("Pushing to the local DSC, Done!")
        resource_id = dataset["offer_iri"]
    local_connector_resource_api.add_contract_to_resource(resource=resource_id,

                                                          contract=contract_id)
    log.debug("Contract added to resource")
    log.info("Sending resource to broker")
    bs = local_connector.send_resource_to_broker(resource_uri=resource_id)
    log.info("Resource was sent to broker.")
    create_created_contract_activity(context, dataset["id"])
    return {"broker success": bs}


@ids_actions.route('/ids/view/publish/<id>', methods=['GET', 'POST'])
def publish(id, offering_info=None, errors=None):
    c = plugins.toolkit.g
    context = {'model': model, 'session': model.Session,
               'user': c.user or c.author, 'auth_user_obj': c.userobj,
               }
    dataset = toolkit.get_action('package_show')(context, {'id': id})
    c.pkg_dict = dataset
    c.usage_policies = config.get("ckanext.ids.usage_control_policies")
    c.offering = {}
    c.errors = {}
    c.current_date_time = datetime.datetime.now(tz=tz.tzlocal()).replace(
        microsecond=0)
    if request.method == "POST":
        try:
            if "multipart/form-data" in str(request.content_type):
                contract = Contract(request.form)
            else:
                contract = Contract(request.get_json())
            c.offering = contract
            c.errors = contract.errors
            if len(contract.errors) == 0:
                request.data = contract
                publish_action(id)
        except KeyError as e:
            log.error(e)

    return toolkit.render('package/publish.html',
                          extra_vars={
                              u'pkg_dict': dataset
                          })


@ids_actions.route('/ids/view/contracts/<id>', methods=['GET'])
def contracts(id, offering_info=None, errors=None):
    local_connector = Connector()
    local_dsc_api = local_connector.get_resource_api()
    c = plugins.toolkit.g
    context = {'model': model, 'session': model.Session,
               'user': c.user or c.author, 'auth_user_obj': c.userobj,
               }
    dataset = toolkit.get_action('package_show')(context, {'id': id})
    c.contracts = []
    contracts = local_dsc_api.get_contracts(dataset["offer_iri"])
    c.contracts = contracts["_embedded"]["contracts"]
    c.data = dataset
    return toolkit.render('package/contracts.html',
                          extra_vars={
                              u'pkg_dict': dataset
                          })

#dtheiler start
@trusts_recommender.route('/ids/actions/store_download_interaction', methods=['POST'])
def store_download_interaction():

    data = clean_dict(dict_fns.unflatten(tuplize_dict(parse_params(request.form))))

    recomm_store_download_interaction(data['entityId'])

    return "true"

@trusts_recommender.route('/ids/actions/store_view_recomm_interaction', methods=['POST'])
def store_view_recomm_interaction():

    data = clean_dict(dict_fns.unflatten(tuplize_dict(parse_params(request.form))))

    recomm_store_view_recomm_interaction(data['entityId'], data['recoId'])

    return "true"
#dtheiler end

# endpoint to accept a contract offer
@ids_actions.route('/ids/actions/contract_accept', methods=['POST'])
def contract_accept():
    """
    Expected body:
    {
        "provider_url" : "ht...",
        "resourceId"   : "ht...",
        "artifactId"   : "ht...",
        "contractId"   : "htt..."
        "brokerResourceId" : "htt....."
    }
    """

    data = clean_dict(
        dict_fns.unflatten(tuplize_dict(parse_params(request.form))))
    if data["provider_url"] is None or len(data["provider_url"]) < 1:
        providing_base_url = "/".join(data["resourceId"].split("/")[:3])
        data["provider_url"] = providing_base_url

    local_connector = Connector()
    local_dsc_api = local_connector.get_resource_api()
    # get the description of the contract

    log.debug(":-:-:-:-:- -----  Description Request  ------ "
              "-:-:-:-:-:-: to" + data['provider_url'] + "/api/ids/data")
    contract = local_dsc_api.descriptionRequest(
        data['provider_url'] + "/api/ids/data", data['contractId'])

    graphs = local_connector.ask_broker_for_description(
        element_uri=data["brokerResourceId"])
    remote_artifacts = graphs_to_artifacts(graphs)

    # Attempt at multiplying permission ---------------------------
    allperms = []
    for arti_num, artifact in enumerate(remote_artifacts):
        newperms = copy.deepcopy(contract["ids:permission"])
        for item in newperms:
            item["ids:target"] = artifact
            item["@id"] += "_" + str(arti_num)
        allperms += newperms

    artifactparm = remote_artifacts
    permparam = allperms
    resource_param = [data['resourceId'] for _ in artifactparm]

    try:
        agreement_response = local_dsc_api.contractRequest(
            data['provider_url'] + "/api/ids/data",
            resource_param,
            artifactparm,  # data['artifactId'],
            False,
            permparam)  # obj)

        local_resource = IdsResource.get(data['resourceId'])
        if local_resource == None:
            local_resource = IdsResource(data['resourceId'])
            local_resource.save()

        local_agreement_uri = agreement_response["_links"]["self"]["href"]
        local_agreement = IdsAgreement(id=local_agreement_uri,
                                       resource=local_resource,
                                       user="admin")
        local_agreement.save()

        log.debug("agreement_uri :\t" + local_agreement_uri)

        local_artifacts = \
            local_dsc_api.get_artifacts_for_agreement(
                local_agreement_uri)
        first_artifact = \
            local_artifacts["_embedded"]["artifacts"][0]["_links"]["self"][
                "href"]
        log.debug("artifact_uri :\t" + first_artifact)

    except IOError as e:
        log.error(e)
        base.abort(500, e)

    #        data_response = local_connector_resource_api.get_data(first_artifact)
    #        size_data = len(data_response.content)
    #        log.debug("size_of_data :\t" + str(size_data))
    #        log.debug("\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
    #        log.debug("\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
    #        log.debug("\n\n\n\n\n")

    #dtheiler start
    if trusts_recommender_plugin_name in toolkit.g.plugins:
        recomm_store_accept_contract_interaction(data['resourceId'])
    #dtheiler end

    if trusts_blockchain_plugin_name in toolkit.g.plugins:
        # stefan_gindl start
        dict_smart_contracts = {
            "channel": "mychannel",
            "msp": "Org1MSP",
            "orguid": "Org1_appuser",
            "assetid": data['resourceId'],
            "target": config.get("ckanext.ids.connector_catalog_iri"),
        }
        r = send_request_to_smart_contract_api(
            'basic_transfer_asset',
            dict_smart_contracts,
        )
        msg_smart_contracts = r.json()['message']
        print("=============== SMART CONTRACT::TR OUTPUT START ==================")
        print(msg_smart_contracts)
        print("=============== SMART CONTRACT OUTPUT END ==================")
        # stefan_gindl end

    return agreement_response
    # resource = local_connector_resource_api.descriptionRequest(data['provider_url'] + "/api/ids/data", data['resourceId'])
    # package = broker_client._to_ckan_package(resource)

    # create_external_package(package)

    # in case of success create the resource on local and add the agreement and other meta on extra



# endpoint to accept a contract offer
@ids_actions.route('/ids/actions/get_data', methods=['GET'])
def get_data():

    local_connector = Connector()
    local_connector_resource_api = local_connector.get_resource_api()

    url = local_connector_resource_api.recipient + "/api/artifacts/" + request.args.get(
        "artifact_id")
    ## get filename
    filename = local_connector_resource_api.get_artifact(url)["title"]
    data_response = local_connector_resource_api.get_data(url)
    response = Response(
        stream_with_context(data_response.iter_content(chunk_size=1024)),
        content_type=data_response.headers.get("Content-Type"),
        status=data_response.status_code
    )
    response.headers["Content-Disposition"] = "attachment;filename="+filename
    return response

# endpoint to create a subscription
@ids_actions.route('/ids/actions/subscribe', methods=['GET'])
def subscribe():
    g = plugins.toolkit.g
    offer_url= request.args.get("offer_url")
    subscriber_email= g.userobj.email

    local_connector = Connector()

    graphs = local_connector.ask_broker_for_description(
        element_uri=offer_url)

    #  This is failing
    dataset = graphs_to_ckan_result_format(graphs)
    # Here we get the local agreements, if they exist
    # --------------------------------------------------------------------
    resourceId = dataset["id"]
    log.info("Getting ..." + resourceId)
    local_resource = IdsResource.get(resourceId)
    if local_resource is None:
        h.flash_error(
            _('An aggreement does not exist yet. Please accept a contract first.'))
        return toolkit.redirect_to('/ids/processExternal?uri=' + offer_url)
    else:
        try:
            log.info("Getting local agreements...")
            local_agreements = local_resource.get_agreements()
            log.info("Creating subscription object...")
            subscription = Subscription(resourceId, local_agreements[0], subscriber_email)
            log.info("Creating subscription...")
            subscription.subscribe()
            log.info("Subscription created successfully!!")
            response = {"offer_url": offer_url, "subscriber_email": subscriber_email}
        except AttributeError as error:
            log.error("There was an error while creating the subscription")
            local_agreements = []
            response = error.with_traceback()
        return response

@ids_actions.route('/ids/actions/unsubscribe', methods=['GET'])
def unsubscribe():
    subscription_url = request.args.get("subscription_url")
    subscription = IdsSubscription.get(subscription_url)
    subscription.delete()


def create_external_package(data):
    # get clean data from the form, data will hold the common meta for all resources

    # iterate through all resources and add them on the package
    # for resource in resources:
    #    toolkit.get_action('resource_create')(None, resource)

    c = plugins.toolkit.g
    context = {'model': model, 'session': model.Session,
               'user': c.user or c.author, 'auth_user_obj': c.userobj,
               }

    try:
        toolkit.get_action("package_create")(context, data)
    # unlikely to happen, but just to demonstrate error handling
    except (ValidationError):
        return base.abort(404, _(u'There was some error on the send data'))
    # redirect tp dataset read page
    return toolkit.redirect_to('dataset.read',
                               id=id)


def create_or_get_catalog_id():
    local_connector_resource_api = Connector().get_resource_api()
    title = config.get("ckanext.ids.local_node_name")
    catalogs = local_connector_resource_api.get_catalogs()
    found = False
    for i, value in enumerate(catalogs["_embedded"]["catalogs"]):
        if value["title"] == title:
            found = True
            catalog_iri = value["_links"]["self"]["href"]
            continue

    if not found:
        catalog = {"title": title}
        catalog_iri = local_connector_resource_api.create_catalog(catalog)
    config.store.update({"ckanext.ids.connector_catalog_iri": catalog_iri})


def merge_extras(old, new):
    # Using defaultdict
    temp = defaultdict(dict)
    log.info("merging extras")
    for elem in old:
        temp[elem['key']] = (elem['value'])
    for elem in new:
        temp[elem['key']] = (elem['value'])
    merged = [{"key": key, "value": value} for key, value in temp.items()]
    return merged


@ids_actions.route('/ids/processExternal', methods=[
    'GET'])
def contracts_remote():
    """
    External resources should have as URL this endpoint
    """

    c = plugins.toolkit.g
    context = {'model': model, 'session': model.Session,
               'user': c.user or c.author, 'auth_user_obj': c.userobj,
               }

    # ToDo For now it assumes the DSC is accessible in the same hostname
    # as the CKAN, but with port 8282
    _dscbaseurl = config.get("ckan.site_url")
    _dsc_hostname = urlsplit(_dscbaseurl).hostname.split(":")[0]
    basedscurl = _dsc_hostname + ":8282"
    log.error("-:-:-:-:~~~~~~~~~~~~~~~~------------------------------>\n\n\n")

    # Ger from broker info for this ID
    resource_uri = request.args.get("uri")
    local_connector = Connector()

    graphs = local_connector.ask_broker_for_description(
        element_uri=resource_uri)

    #  This is failing
    dataset = graphs_to_ckan_result_format(graphs)

    c.pkg_dict = dataset
    contracts = graphs_to_contracts(graphs,
                                    broker_resource_uri=resource_uri)

    c.contracts = contracts

    # Here we get the local agreements, if they exist
    # --------------------------------------------------------------------
    resourceId = dataset["id"]
    local_resource = IdsResource.get(resourceId)
    # log.debug(json.dumps(dataset,indent=1))

    if trusts_recommender_plugin_name in toolkit.g.plugins:
        #dtheiler start
        recomm_store_view_interaction(
            dataset["id"], #entityId
            dataset["type"]) #entityType
        #dtheiler end

    try:
        local_agreements = local_resource.get_agreements()
    except AttributeError:
        local_agreements = []

    local_dsc_API = local_connector.get_resource_api()
    local_artifacts = []
    for local_agreement in local_agreements:
        if local_agreement is not None:
            site_url = str(toolkit.config.get('ckan.site_url'))

            artifacts = local_dsc_API.get_artifacts_for_agreement(
                local_agreement.id)
            log.debug("~~~~~~~~~~~\n|~\n|~\n|~")
            log.debug("\tagreement_uri: \t" + local_agreement.id)
            if "_embedded" in artifacts.keys():
                for ar in artifacts["_embedded"]["artifacts"]:

                    artifacturi = ar["_links"]["self"]["href"]

                    artifactuuid = artifacturi.split("/")[-1]
                    arttitle = ar["title"] if len(
                        ar["title"]) > 0 else artifactuuid
                    artdesc = ar["description"]

                    if "service_base_access_url" in arttitle:
                        url = basedscurl + "/api/artifacts/" + artifactuuid
                        accessurl = url + "/data"
                    else:
                        accessurl = \
                            site_url + "/ids/actions/get_data?artifact_id=" \
                                       "" + artifactuuid

                    # log.debug("\t|\n\t\n\t|~~~~~~~~~~~>")
                    # log.debug("\t\tartifact_uri :\t" + artifacturi)
                    # log.debug("\t\taccessurl: \t" + accessurl)
                    # log.debug("\t\ttitle: \t" + arttitle)
                    # log.debug("\t\tdescription: \t" + artdesc)

                    artifact_description = {"url": accessurl}
                    artifact_description["title"] = arttitle
                    artifact_description["description"] = artdesc
                    if "title" in ar.keys() and len(ar["title"]) > 0:
                        artifact_description["title"] = ar["title"]
                    if "description" in ar.keys() and len(
                            ar["description"]) > 0:
                        artifact_description["description"] = ar["description"]

                    local_artifacts.append(artifact_description)

    if len(local_artifacts):
        c.local_artifacts = local_artifacts
    c.data = dataset
    return toolkit.render('package/contracts_external.html',
                          extra_vars={
                              u'pkg_dict': dataset
                          })



from flask_swagger_ui import get_swaggerui_blueprint

SWAGGER_URL = '/api/docs'  # URL for exposing Swagger UI (without trailing '/')
API_URL = '/api/swagger/ckan_ids_openapi.yaml'  # Our API url (can of course be a local resource)

# Call factory function to create our blueprint
swaggerui_blueprint = get_swaggerui_blueprint(
    SWAGGER_URL,  # Swagger UI static files will be mapped to '{SWAGGER_URL}/dist/'
    API_URL,
    config={  # Swagger UI config overrides
        'app_name': "TRUSTS CKAN API Documentation"
    },
    # oauth_config={  # OAuth config. See https://github.com/swagger-api/swagger-ui#oauth2-configuration .
    #    'clientId': "your-client-id",
    #    'clientSecret': "your-client-secret-if-required",
    #    'realm': "your-realms",
    #    'appName': "your-app-name",
    #    'scopeSeparator': " ",
    #    'additionalQueryStringParams': {'test': "hello"}
    # }
)



import logging
import datetime

from sqlalchemy import Table
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import types
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.orm import relation

from ckan.model.meta import metadata, mapper, Session
from ckan.model.domain_object import DomainObject

log = logging.getLogger(__name__)

__all__ = [
    'IdsAgreement', 'ids_agreement_table',
    'IdsResource', 'ids_resource_table',
    'IdsSubscription', 'ids_subscription_table'
]

ids_agreement_table = None
ids_resource_table = None
ids_subscription_table = None


def setup():
    if ids_subscription_table is None:
        define_ids_tables()
        log.debug('IDS tables defined in memory')

    if not ids_agreement_table.exists():

        # Create each table individually rather than
        # using metadata.create_all()
        ids_resource_table.create()
        ids_agreement_table.create()
        ids_subscription_table.create()
        log.debug("IDS tables created")
    elif not ids_subscription_table.exists():
        ids_subscription_table.create()
        log.debug("IDS subscription table added.")
        pass
    else:
        from ckan.model.meta import engine
        log.debug('IDS tables already exist')
        # Check if existing tables need to be updated
        inspector = Inspector.from_engine(engine)


class IdsDomainObject(DomainObject):
    '''Convenience methods for searching objects
    '''
    key_attr = 'id'

    @classmethod
    def get(cls, key, default=None, attr=None):
        '''Finds a single entity in the register.'''
        if attr is None:
            attr = cls.key_attr
        kwds = {attr: key}
        o = cls.filter(**kwds).first()
        if o:
            return o
        else:
            return default

    @classmethod
    def filter(cls, **kwds):
        query = Session.query(cls).autoflush(False)
        return query.filter_by(**kwds)


class IdsResource(IdsDomainObject):
    '''An IDS Resource
    '''
    def __repr__(self):
        return '<IdsResource id=%s>' % \
               (self.id)

    def __str__(self):
        return self.__repr__().encode('ascii', 'ignore')

    def __init__(self, id=None):
        self.id = id

    def get_agreements(self, user_id=None):
        """ get the agreements for this resource """

        query = Session.query(IdsAgreement).filter(IdsAgreement.resource_id == self.id)

        if user_id is not None:
            query = query.filter(IdsAgreement.user_id == user_id)

        return query.all()


class IdsAgreement(IdsDomainObject):
    '''An agreement that has been made between a consumer and a provider data space connector
    '''
    def __repr__(self):
        return '<IdsAgreement id=%s resource_id=%s user_id=%s>' % \
               (self.id, self.resource_id, self.agreement_id, self.user_id)

    def __str__(self):
        return self.__repr__().encode('ascii', 'ignore')

    def __init__(self, id=None, resource=None, user=None):
        self.id = id
        self.resource_id = resource.id
        self.user_id = user

    def get_subscriptions(self, user_id=None):
        """ get the agreements for this agreement """

        query = Session.query(IdsSubscription).filter(IdsSubscription.agreement_id == self.id)

        if user_id is not None:
            query = query.filter(IdsSubscription.user_id == user_id)

        return query.all()


class IdsSubscription(IdsDomainObject):
    '''A subscription that has been made between a consumer and a provider data space connector
    '''
    def __repr__(self):
        return '<IdsSubscription id=%s agreement_id=%s user_id=%s>' % \
               (self.id, self.agreement_id, self.user_id)

    def __str__(self):
        return self.__repr__().encode('ascii', 'ignore')

    def __init__(self, id=None, agreement=None, user=None):
        self.id = id
        self.agreement_id = agreement.id
        self.user_id = user


def define_ids_tables():

    global ids_agreement_table
    global ids_resource_table
    global ids_subscription_table

    ids_resource_table = Table(
        'ids_resource',
        metadata,
        Column('id', types.UnicodeText, primary_key=True),
        Column('created', types.DateTime, default=datetime.datetime.utcnow)
    )

    ids_agreement_table = Table(
        'ids_agreement',
        metadata,
        Column('id', types.UnicodeText, primary_key=True),
        Column('resource_id', types.UnicodeText, ForeignKey('ids_resource.id')),
        Column('created', types.DateTime, default=datetime.datetime.utcnow)
    )

    ids_subscription_table = Table(
        'ids_subscription',
        metadata,
        Column('id', types.UnicodeText, primary_key=True),
        Column('agreement_id', types.UnicodeText, ForeignKey('ids_agreement.id')),
        Column('created', types.DateTime, default=datetime.datetime.utcnow)
    )

    mapper(
        IdsResource,
        ids_resource_table,
        properties={
            'agreements': relation(
                IdsAgreement,
                lazy=True,
                backref=u'resource',
            )}
    )

    mapper(
        IdsAgreement,
        ids_agreement_table,
        properties={
            'subscriptions': relation(
                IdsSubscription,
                lazy=True,
                backref=u'subscription',
            )}
    )

    mapper(IdsSubscription,
           ids_subscription_table,
           )


class Resource(dict):
    representation_iri: None
    artifact_iri: None
    service_accessURL: None
    title: None
    description: None
    mediaType: None
    filename: None

    def __init__(self, resource_dict):
        self.service_accessURL = None
        self.representation_iri = None
        self.artifact_iri = None
        self.title = ""
        self.description = ""
        self.mediaType = ""

        if "representation" in resource_dict:
            self.representation_iri = resource_dict["representation"]
        if "artifact" in resource_dict:
            self.artifact_iri = resource_dict["artifact"]
        if 'service_accessURL' in resource_dict:
            self.service_accessURL = resource_dict["service_accessURL"]
        if 'description' in resource_dict:
            self.description = resource_dict['description']
        if 'mimetype' in resource_dict:
            self.mediaType = resource_dict['mimetype']

        # The artifact's title can be taken from one of several
        if 'name' in resource_dict:
            self.title = resource_dict["name"]
        if 'title' in resource_dict:
            self.title = resource_dict["title"]
        if 'resource_type' in resource_dict and resource_dict["resource_type"]:
            self.title = resource_dict["resource_type"] +"__"+ self.title






import json
import os
import requests


URL_SUBDIRECTORIES = {
    'basic_create_asset': 'basic/createasset',
    'basic_get_all_assets': 'basic/getallassets',
    'basic_init_ledger': 'basic/initledger',
    'basic_transfer_asset': 'basic/transferasset',
    'basic_update_asset': 'basic/updateasset',
    'scoring_get_all_scores': 'scoring/getallscores',
    'scoring_init_ledger': 'scoring/initledger',
    'scoring_update': 'scoring/updatescore',
    'secure_agree_to_buy': 'secure/agreetobuy',
    'secure_agree_to_sell': 'secure/agreetosell',
    'secure_create_asset': 'secure/createasset',
    'secure_read_asset': 'secure/readasset',
    'secure_read_bid_price': 'secure/readbidprice',
    'secure_read_sale_price': 'secure/readsaleprice',
    'secure_transfer_asset': 'secure/transferasset',
    'secure_update_description': 'secure/updatedescription',
    'secure_verify_asset': 'secure/verifyasset',
    }


def make_request_to_smart_contract_api(base_url):

    headers = {'Content-Type': 'application/json'}

    def __create_urls(base_url):
        return {method_name: os.path.join(base_url, url_subdirectory)
                for method_name, url_subdirectory
                in URL_SUBDIRECTORIES.items()}

    urls = __create_urls(base_url)

    def send_request_to_smart_contract_api(kind, data):
        return requests.get(urls[kind],
                            headers=headers,
                            data=json.dumps(data))

    return send_request_to_smart_contract_api



"""
The Offer Model of the Dataspace Connector
{
    "title": "string",
    "description": "string",
    "keywords": [
        "string"
    ],
    "publisher": "string",
    "language": "string",
    "license": "string",
    "sovereign": "string",
    "endpointDocumentation": "string",
    "samples": [
        "string"
    ],
    "paymentMethod": "undefined"
}
"""

import logging
import json

class Offer:

    def __init__(self, pkg_dict):
        logging.debug("\n\n\n---------------------------- CREATING OFFER")
        logging.debug(json.dumps(pkg_dict, indent=1))
        self.access_url = None
        self.license =None
        self.title = pkg_dict['title']
        self.keywords = pkg_dict['tags']
        self.publisher = pkg_dict['owner_org']
        self.description = pkg_dict['description']
        additional = {
            "https://www.trusts-data.eu/ontology/asset_type":
                "https://www.trusts-data.eu/ontology/" + str(pkg_dict[
                                                          'type']).capitalize(),
            "https://www.trusts-data.eu/ontology/theme": pkg_dict['theme']
        }
        self.additional = additional
        if "base_URL" in pkg_dict.keys():
            self.access_url = pkg_dict["base_URL"]

        if "license_url" in pkg_dict.keys():
            self.license = pkg_dict["license_url"]

        if "offer_iri" in pkg_dict.keys():
            self.offer_iri = pkg_dict["offer_iri"] if pkg_dict["offer_iri"] != "" else None
        else:
            self.offer_iri = None

        if "catalog_iri" in pkg_dict.keys():
            self.catalog_iri = pkg_dict["catalog_iri"] if pkg_dict["catalog_iri"] != "" else None
        else:
            self.catalog_iri = None

        logging.debug("\n----------------------------OFFER|")


    def to_dictionary(self):
        d = {
            'title': self.title,
            'keywords': self.keywords,
            'publisher': self.publisher,
            'license': self.license,
            'description': self.description,
        #    'catalog_iri': self.catalog_iri,
        #    'offer_iri': self.offer_iri
        }
        for k,v in self.additional.items():
            if k.startswith("ids:"):
                d[k.split(":")[-1]] = v
            else:
                d[k] = v

        return d



"""
{
  "author": null, 
  "author_email": null, 
  "creator_user_id": "60cf572d-5de0-4ef5-b661-51bfd332f3c2", 
  "extras": [
    {
      "key": "artifact", 
      "value": "http://localhost:8089/api/artifacts/4797ba8e-b6e6-470a-93a0-84f97432ff20"
    }, 
    {
      "key": "catalog", 
      "value": "http://localhost:8089/api/catalogs/f6920af2-9697-48b6-aa8d-08f62956bf89"
    }, 
    {
      "key": "offers", 
      "value": "http://localhost:8089/api/offers/2fdc1dc4-d7fa-4164-841b-8ccea0d8a6f7"
    }, 
    {
      "key": "representation", 
      "value": "http://localhost:8089/api/representations/ddf5187b-0710-4eb7-a013-b4044e435ac2"
    }
  ], 
  "groups": [], 
  "id": "24b10e55-e01d-48a1-894a-c576ff532b15", 
  "isopen": true, 
  "license_id": "cc-by", 
  "license_title": "Creative Commons Attribution", 
  "license_url": "http://www.opendefinition.org/licenses/cc-by", 
  "maintainer": null, 
  "maintainer_email": null, 
  "metadata_created": "2021-11-25T16:43:17.205565", 
  "metadata_modified": "2021-11-30T13:42:23.994829", 
  "name": "test2", 
  "notes": null, 
  "num_resources": 3, 
  "num_tags": 0, 
  "organization": {
    "approval_status": "approved", 
    "created": "2021-11-24T17:19:04.644099", 
    "description": "", 
    "id": "e5064645-2841-4a37-bb48-279f68c5bc2b", 
    "image_url": "", 
    "is_organization": true, 
    "name": "test", 
    "state": "active", 
    "title": "test", 
    "type": "organization"
  }, 
  "owner_org": "e5064645-2841-4a37-bb48-279f68c5bc2b", 
  "private": false, 
  "relationships_as_object": [], 
  "relationships_as_subject": [], 
  "resources": [
    {
      "cache_last_updated": null, 
      "cache_url": null, 
      "configurationdatetime_date": "", 
      "configurationdatetime_time": "", 
      "created": "2021-11-25T16:43:47.682071", 
      "deploymentdatetime_date": "", 
      "deploymentdatetime_time": "", 
      "description": null, 
      "format": "", 
      "hash": "", 
      "id": "6caffd25-c361-4b6f-b1c6-89f03bf313b1", 
      "last_modified": null, 
      "metadata_modified": "2021-11-25T16:43:47.669803", 
      "mimetype": null, 
      "mimetype_inner": null, 
      "name": "", 
      "package_id": "24b10e55-e01d-48a1-894a-c576ff532b15", 
      "position": 0, 
      "readmedatetime_date": "", 
      "readmedatetime_time": "", 
      "resource_type": "deploymentPhoto", 
      "size": null, 
      "state": "active", 
      "url": "", 
      "url_type": null
    }
  ], 
  "state": "active", 
  "tags": [], 
  "theme": "https://trusts.org/vacabulary/themes/Health", 
  "title": "test", 
  "type": "dataset", 
  "url": null, 
  "version": null
}
"""


#!/usr/bin/env python3
#
# Copyright 2020 Fraunhofer Institute for Software and Systems Engineering
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import requests

# Suppress ssl verification warning
requests.packages.urllib3.disable_warnings()


class SubscriptionApi:
    session = None
    recipient = None

    def __init__(self, recipient, auth=("admin", "password")):
        self.session = requests.Session()
        self.session.auth = auth
        self.session.verify = False

        self.recipient = recipient

    def create_subscription(self, data={}):
        response = self.session.post(self.recipient + "/api/subscriptions", json=data)
        return response.headers["Location"]

    def subscription_message(self, data={}, params={}):
        response = self.session.post(
            self.recipient + "/api/ids/subscribe", json=data, params=params
        )
        return response

    def get_subscriptions(self):
        response = self.session.get(self.recipient + "/api/subscriptions")
        return response


from six.moves.urllib.parse import urlparse
from ckan.logic import validators
import string
from ckan.common import _
from ckanext.scheming.validation import register_validator, scheming_validator

@register_validator
def trusts_url_validator(key, data, errors, context):
    ''' Checks that the provided value (if it is present) is a valid URL '''

    url = data.get(key, None)
    if not url:
        return

    try:
        pieces = urlparse(url)
        if all([pieces.scheme, pieces.netloc]) and \
                set(pieces.netloc) <= set(string.ascii_letters + string.digits + '-.:') and \
                pieces.scheme in ['http', 'https']:
            return
    except ValueError:
        # url is invalid
        pass

    errors[key].append(_('Please provide a valid URL'))

    # Custom Activities

_object_id_validators = {
    'pushed to dataspace connector': validators.package_id_exists,
    'created contract': validators.package_id_exists
}

def object_id_validator(key, activity_dict, errors, context):
    '''Validate the 'object_id' value of an activity_dict.
    This wraps ckan.logic.validators.object_id_validator to support additional
    object types
    '''
    activity_type = activity_dict[('activity_type',)]
    if activity_type in _object_id_validators:
        object_id = activity_dict[('object_id',)]
        return _object_id_validators[activity_type](object_id, context)
    return validators.object_id_validator(key, activity_dict, errors, context)


def activity_type_exists(activity_type):
    '''Wrap ckan.logic.validators.activity_type_exists to support additional
    activity stream types
    '''
    if activity_type in _object_id_validators:
        return activity_type
    return validators.activity_type_exists(activity_type)


import logging
import json
from ckan.common import is_flask_request

from ckan.common import _, c, request

log = logging.getLogger(__name__)


def check_if_contract_offer_exists(id):
    return True

def string_to_json(json_string):
    return json.loads(json_string)

def get_facet_items_dict(
        facet, search_facets=None, limit=None, exclude_active=False):
    '''Return the list of unselected facet items for the given facet, sorted
    by count.
    Returns the list of unselected facet contraints or facet items (e.g. tag
    names like "russian" or "tolstoy") for the given search facet (e.g.
    "tags"), sorted by facet item count (i.e. the number of search results that
    match each facet item).
    Reads the complete list of facet items for the given facet from
    c.search_facets, and filters out the facet items that the user has already
    selected.
    Arguments:
    facet -- the name of the facet to filter.
    search_facets -- dict with search facets(c.search_facets in Pylons)
    limit -- the max. number of facet items to return.
    exclude_active -- only return unselected facets.
    '''
    if search_facets is None:
        search_facets = getattr(c, u'search_facets', None)

    if not search_facets or not search_facets.get(
            facet, {}).get('items'):
        return []
    facets = []
    for facet_item in search_facets.get(facet)['items']:
        if not len(facet_item['name'].strip()):
            continue
        params_items = request.params.items(multi=True) \
            if is_flask_request() else request.params.items()
        if not (facet, facet_item['name']) in params_items:
            facets.append(dict(active=False, **facet_item))
        elif not exclude_active:
            facets.append(dict(active=True, **facet_item))
    # Sort descendingly by count and ascendingly by case-sensitive display name
    facets.sort(key=lambda it: (-it['count'], it['display_name'].lower()))
    if hasattr(c, 'search_facets_limits'):
        if c.search_facets_limits and limit is None:
            limit = c.search_facets_limits.get(facet)
    # zero treated as infinite for hysterical raisins
    if limit is not None and limit > 0:
        return facets[:limit]
    return facets


def has_more_facets(facet, search_facets, limit=None, exclude_active=False):
    '''
    Returns True if there are more facet items for the given facet than the
    limit.
    Reads the complete list of facet items for the given facet from
    c.search_facets, and filters out the facet items that the user has already
    selected.
    Arguments:
    facet -- the name of the facet to filter.
    search_facets -- dict with search facets(c.search_facets in Pylons)
    limit -- the max. number of facet items.
    exclude_active -- only return unselected facets.
    '''
    facets = []
    for facet_item in search_facets.get(facet)['items']:
        if not len(facet_item['name'].strip()):
            continue
        params_items = request.params.items(multi=True) \
            if is_flask_request() else request.params.items()
        if not (facet, facet_item['name']) in params_items:
            facets.append(dict(active=False, **facet_item))
        elif not exclude_active:
            facets.append(dict(active=True, **facet_item))
    if hasattr(c, 'search_facets_limits'):
        if c.search_facets_limits and limit is None:
            limit = c.search_facets_limits.get(facet)
    if limit is not None and len(facets) > limit:
        return True
    return False